"""Contains :class:`.fkey.ForeignKey`."""

from typing import Optional, Type

from .exceptions import IncompatibleArgument

CASCADE = "cascade"


class ForeignKey:
    """Representer of foreign key relations.

    This class implements logic handling relations across tables through connector columns.

    :param foreign_model:   `MappedClass` that represents a foreign model
    :param reverse_name:    keyword-only. Relation name on the related class tied to this class
    :param on_delete:       keyword-only. Whether to delete cascade style or not. Optional
    :param on_update:       keyword-only. Whether to update cascade style or not. Optional.
    """

    def __init__(
        self,
        foreign_model: Type,
        *,
        reverse_name: str,
        on_delete: Optional[str] = None,
        on_update: Optional[str] = None,
    ) -> None:
        self.foreign_model = foreign_model
        self.reverse_name = reverse_name
        self.on_delete = on_delete
        self.on_update = on_update
        if self.on_delete not in [None, CASCADE]:
            raise IncompatibleArgument("Wrong value for `on_delete`")
        if self.on_update not in [None, CASCADE]:
            raise IncompatibleArgument("Wrong value for `on_update`")



"""Contains column-like classes."""

from __future__ import annotations

from typing import Any, Dict, Iterable, List, Optional, Type

from .exceptions import IncompatibleArgument, OrphanColumn
from .fkey import ForeignKey
from .types import Serial, String


class BaseColumn:
    """Base class for representing a column-like value in a database."""

    aggs: List[str] = []

    def __init__(self) -> None:
        self.alias = None

    def binary_ops(self, other: Any, operator: str) -> str:
        """String representation for direct Python binary operations between :class:`column.BaseColumn` objects.

        E.g.::

            Order.price <= 200.00

        or

            u = User(fullname="Abigail Smith", age=30)
            User.age > u.age

        Has the following:

        :param other:       a literal or a :class:`column.BaseColumn` object, to compare with this object
        :param operator:    SQL operator, represented as a string.
        """
        other_compound = "%s.%s" % (other.view_name, other.variable_name) if isinstance(other, BaseColumn) else other
        full_field_name = ".".join(self.compound_variable_name.split(", "))
        return "%s %s %s" % (full_field_name, operator, other_compound)

    def __eq__(self, other: Any) -> str:
        return self.binary_ops(other, "=")

    def __ne__(self, other: Any) -> str:
        return self.binary_ops(other, "<>")

    def __lt__(self, other: Any) -> str:
        return self.binary_ops(other, "<")

    def __gt__(self, other: Any) -> str:
        return self.binary_ops(other, ">")

    def __le__(self, other: Any) -> str:
        return self.binary_ops(other, "<=")

    def __ge__(self, other: Any) -> str:
        return self.binary_ops(other, ">=")

    def like(self, other: str) -> str:
        return self.binary_ops(other, "LIKE")

    def in_(self, flat_list: Iterable[Any]) -> str:
        flat_list_str = "(%s)" % flat_list
        return self.binary_ops(flat_list_str, "IN")

    def not_in(self, flat_list: Iterable[Any]) -> str:
        flat_list_str = "(%s)" % flat_list
        return self.binary_ops(flat_list_str, "NOT IN")

    def label(self, alias: str) -> BaseColumn:
        """Returns the ORM representation for SQL column aliasing.

        :param alias:   alias as a string.
        """
        self.alias = alias
        return self


class Scalar(BaseColumn):
    """Scalar value as a column.

    :param repr_:   SQL representation of the scalar name.
    """

    def __init__(self, repr_) -> None:
        super(Scalar, self).__init__()
        self.compound_variable_name = repr_


class Field(BaseColumn):
    """Representer of all real and virtual fields."""

    @property
    def compound_variable_name(self):
        return "%s, %s" % (self.view_name, self.variable_name)


class Column(Field):
    """Abstraction of a real table column in a database.

    :param type_:       type of value that this column expects. Must be one of the types defined in :module:`.types`
    :param length:      max length of the expected value. Only works with :class:`.types.String`. Optional
    :param rel:         marker of a relationship -- a foreign key. Optional
    :param primary_key: keyword-only. Whether or not the column is a primary key. Optional
    :param default:     keyword-only. Default value for cells of the column to take. Optional
    :param nullable:    keyword-only. Whether or not the cells of the column are nullable. Optional.
    """

    objects: Dict[int, Dict[str, Any]] = {}

    def __init__(
        self,
        type_: Type,
        length: Optional[int] = None,
        rel: Optional[ForeignKey] = None,
        *,
        primary_key: bool = False,
        default: Any = None,
        nullable: bool = True,
    ) -> None:
        super(Column, self).__init__()
        self.type = type_
        self.length = length
        self.rel = rel
        self.primary_key = primary_key
        self.default = default
        self.nullable = nullable
        if self.primary_key and self.type not in [Serial, String]:
            raise IncompatibleArgument("Wrong type for primary key.")

        if self.rel is not None:
            if not isinstance(self.rel, ForeignKey):
                raise IncompatibleArgument("Relationship should be a `ForeignKey`.")

            if self.primary_key or self.default or not self.nullable:
                raise IncompatibleArgument("Wrong combination of arguments with `ForeignKey`.")

            if self.type not in [Serial, String]:
                raise IncompatibleArgument("Wrong type for `ForeignKey`.")

        if self.length is not None:
            if self.type != String:
                raise IncompatibleArgument("Only `String` type can have length.")

            if not isinstance(self.length, int):
                raise IncompatibleArgument("`String` type should have an integer length.")

    @property
    def model(self) -> Type:
        """Relational model that the column belongs to."""
        try:
            m = self.objects[id(self)]["model"]
        except KeyError:
            raise OrphanColumn
        return m

    @property
    def variable_name(self) -> str:
        """Name with which the column is defined."""
        try:
            v = self.objects[id(self)]["variable_name"]
        except KeyError:
            raise OrphanColumn
        return v

    @property
    def view_name(self) -> str:
        """Name of the SQL table that the column belongs to."""
        return self.model.get_table_name()


class VirtualField(Field):
    """Abstraction of a virtual table column in a database.

    Never instantiated directly but is derived from actual columns, e.g. by accessing a field of a
    :class:`.query.Subquery` object.

    :param variable_name:   name, as a string, of the column as it is defined in the source
    :param view_name:       name of the view in which the column belongs.
    """

    def __init__(self, variable_name: str, view_name: str) -> None:
        super(VirtualField, self).__init__()
        self.variable_name = variable_name
        self.view_name = view_name



"""Contains :class:`.query.Query` and :class:`.query.Subquery`.

Also contains the class for complete and incomplete row-like objects, namely, :class:`.query.Record`, and the class for
any collection of them :class:`.query.QuerySet`.
"""

from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional, Type, Union

import pyodbc

from .column import BaseColumn, Column, VirtualField
from .db_engine import AbstractEngine
from .exceptions import EntityError, FieldNotExist, MethodChainingError, MultipleResultsFound, QueryFormatError


class Record:
    """Representer of each of a database fetch results.

    This can be a complete or an incomplete table row.

    Proxy class, never directly instantiated.

    :param dct:     data that the record is based on
    :param query:   query that fetched this record, among possibly others.
    """

    def __init__(self, dct: Dict[str, Any], query: Query) -> None:
        super().__setattr__("dct", dct)
        super().__setattr__("query", query)

    def __getattr__(self, attr: str) -> Any:
        try:
            return self.dct[attr]
        except KeyError as e:
            if self.is_complete_row:
                self_model = self.query.entities[0]
                for m in self_model.dep_mapping[self_model]:
                    connector = m.get_connector_column(self_model)
                    if connector.rel.reverse_name == attr:
                        condition_dict = {
                            connector.variable_name: getattr(self, self_model.get_primary_key_column().variable_name)
                        }
                        return Query(m).join(self_model).filter_by(**condition_dict).subquery()

            raise FieldNotExist(attr) from e

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr not in self.dct:
            raise FieldNotExist(attr)

        self.query.filter_by(**self.dct).update(**{attr: value})

    @property
    def is_complete_row(self) -> bool:
        """Whether or not this record is a complete row."""
        return len(self.query.entities) == 1 and isinstance(self.query.entities[0], type)


class QuerySet:
    """A class that represents database fetch results.

    Immutable, ordered. Indexable, iterable, subscriptable.

    NOTE that this is terminal: no any query methods can be applied to the instance anymore.

    :param lst: underlying list of records.
    """

    def __init__(self, lst: List[Record]) -> None:
        self.lst = lst

    def __iter__(self) -> Iterator[Record]:
        yield from self.lst

    def __len__(self) -> int:
        return len(self.lst)

    def __getitem__(self, key: int) -> Record:
        return self.lst[key]

    def __setitem__(self, key: int, value: Record) -> None:
        self.lst[key] = value

    def __delitem__(self, key: int) -> None:
        del self.lst[key]

    def __getslice__(self, start: int, stop: int) -> QuerySet:
        return QuerySet(self.lst[start:stop])

    def __setslice__(self, start: int, stop: int, sequence: List[Any]) -> None:
        self.lst[start:stop] = sequence

    def __delslice__(self, start: int, stop: int) -> None:
        del self.lst[start:stop]


class Subquery:
    """Representer of an SQL subquery.

    A subquery is a nested SELECT statement that is used within another SQL statement.

    Never directly instantiated, but rather initialised by invoking :meth:`.query.Query.subquery()`.

    :param inner_sql:       SQL string of the view represented by the subquery
    :param column_names:    original names of the columns in that view.
    """

    subquery_idx = 0

    def __init__(self, inner_sql: str, column_names: List[str]) -> None:
        type(self).subquery_idx += 1
        self.inner_sql = inner_sql
        self.column_names = column_names
        self.view_name = "anon_%d" % self.subquery_idx
        self.full_sql = "(%s) AS %s" % (self.inner_sql, self.view_name)

    def __str__(self) -> str:
        return self.inner_sql

    def __getattr__(self, attr: str) -> Any:
        if attr not in self.column_names:
            raise EntityError("Wrong field for subquery.")

        return VirtualField(attr, self.view_name)


class QueryBuilder:
    """Builder of an SQL query."""

    def __init__(self) -> None:
        self.data: Dict[str, List[Any]] = {}

    def add_to_data(self, key: str, val: str) -> None:
        """Appends a value to the list `self.data[key]`.

        If `key` does not exist in `self.data`, instantiates it as an empty list and append to it.
        """
        self.data.setdefault(key, []).append(val)

    def build(self) -> str:
        """Builds and returns the final SQL query."""

        def parse_item(compound_name: str) -> str:
            parts = compound_name.split(", ")
            if len(parts) == 1:
                return compound_name
            if len(parts) == 2:
                return "%s.%s" % (*parts,)
            if len(parts) == 3:
                return "%s.%s AS %s" % (*parts,)
            incomplete_res = "%s.%s" % (parts[0], parts[1])
            for part in parts[:2:-1]:
                incomplete_res = "%s(%s)" % (part, incomplete_res)
            return "%s AS %s" % (incomplete_res, parts[2])

        parsed_str = ""
        if "select" in self.data:
            table = self.data["from"][0]

            to_remove = "%s, *" % table
            if len(self.data["select"]) > 1 and to_remove in self.data["select"]:
                self.data["select"].remove(to_remove)

            column_seq = ", ".join(parse_item(s) for s in self.data["select"])

            if "distinct" in self.data:
                column_seq = "DISTINCT %s" % column_seq
            parsed_str += "SELECT %s FROM %s" % (column_seq, table)
        elif "delete" in self.data:
            table = self.data["from"][0]
            parsed_str = "DELETE FROM %s" % table
        elif "update" in self.data:
            table = self.data["from"][0]
            fields_values_seq = ", ".join("%s.%s = %s" % (table, field, value) for field, value in self.data["set"])
            parsed_str = "UPDATE %s SET %s" % (table, fields_values_seq)

        if "from_as" in self.data:
            parsed_str += " AS %s" % self.data["from_as"][0]

        if "join" in self.data:
            condition = " AND ".join(expr for expr in self.data["on"])
            parsed_str += " JOIN %s ON %s" % (self.data["join"][0], condition)

        if "where" in self.data:
            condition = " AND ".join(expr for expr in self.data["where"])
            parsed_str += " WHERE %s" % condition

        if "group_by" in self.data:
            column_name_seq = ", ".join(
                name if len(name.split(", ")) == 1 else "%s.%s" % (*name.split(", "),) for name in self.data["group_by"]
            )
            parsed_str += " GROUP BY %s" % column_name_seq

        if "having" in self.data:
            condition = " AND ".join(expr for expr in self.data["having"])
            parsed_str += " HAVING %s" % condition

        if "order_by" in self.data:
            column_name_seq = ", ".join(
                name if len(name.split(", ")) == 1 else "%s.%s" % (*name.split(", "),) for name in self.data["order_by"]
            )
            parsed_str += " ORDER BY %s" % column_name_seq

        if "limit" in self.data:
            parsed_str += " LIMIT %s" % self.data["limit"][0]

        if "offset" in self.data:
            parsed_str += " OFFSET %s" % self.data["offset"][0]

        if "desc" in self.data:
            parsed_str += " DESC"

        return parsed_str


class Query:
    """Main abstraction for querying for the whole ORM.

    There are two ways to generate :class:`.query.Query` objects: either by calling the
    `.db_session.DBSession.query` method, which is the most usual way, or, less commonly, by instantiating
    `.query.Query` directly.

    Gets as an argument:

    :param entities: -- which correspond to the "columns" of the matched results. May contain at most one `MappedClass` instance. NOTE that `MappedClass` is any subclass of :class:`.model.Model`.
    """

    def __init__(self, *entities: Union[Type, BaseColumn]) -> None:
        self.entities = entities
        self.builder = QueryBuilder()
        if not self.entities:
            raise EntityError("No fields specified for querying.")

        if sum(isinstance(item, type) for item in self.entities) > 1:
            raise EntityError("More than one model specified in the query.")

        for item in self.entities:
            if isinstance(item, type):
                self.mapped_class = item
                table_name = item.get_table_name()
                self.builder.add_to_data("select", "%s, *" % table_name)
                self.builder.add_to_data("from", table_name)
                if item.alias is not None:
                    self.builder.add_to_data("from_as", item.alias)
            elif isinstance(item, BaseColumn):
                if item.alias is not None:
                    comp_name = ", ".join([item.compound_variable_name, item.alias, *item.aggs])
                    item.aggs = []
                    self.builder.add_to_data("select", comp_name)
                else:
                    self.builder.add_to_data("select", item.compound_variable_name)

                if isinstance(item, Column):
                    self.builder.add_to_data("from", item.view_name)
            else:
                raise QueryFormatError

    def __str__(self) -> str:
        return self._sql

    @property
    def _sql(self) -> str:
        """Gets the SQL representation of the current query.

        :return: valid SQL string.
        """
        return self.builder.build()

    def join(self, mapped: Union[Type, Subquery], *exprs: Any) -> Query:
        """Joins the mapped to the the current instance.

        :param mapped:  joined entity
        :param exprs:   list of expressions on which to join.

        :return:        resulting :class:`.query.Query` object.

        :param:`mapped` is either another model or a subquery instance. In the first case, one can join with or without
        expressions. And if the other model is joined without expressions, an expression is constructed via its primary
        key. Note that this is only possible when the other model is a parent model to this one.

        E.g.::

            session.query(User, email_address, last_visited).join(Organisation).all()

            SELECT users.email_address, users.last_visited FROM users
            JOIN organisations ON users.org_id = organisations.id;

        If the other model is joined with expressions, those expressions are used.

        E.g.::

            session.query(Sale, category_name, quantity)
                .join(Category, Category.id == Sales.category_id, Sale.quantity > 100, Category.is_current == 1)
                .all()

            SELECT sales.category_name, sales.quantity FROM sales
            JOIN categories
                ON categories.id = sales.category_id AND sales.quantity > 100 AND categories.is_current = 1;

        In case :param:`mapped` is a subquery object, one can only join with expressions.

        E.g.::

            sq = session.query(Human, Human.id, Human.full_name).subquery()
            session.query(Pet, Pet.name, Pet.age).join(sq, sq.full_name == Pet.name)

            SELECT pets.name, pets.age FROM pets
            JOIN
                (SELECT humans.id, humans.full_name FROM humans) AS anon_27
                ON anon_27.full_name = pets.name;
        """
        self_model = self.mapped_class or self.entities[0].model
        if isinstance(mapped, type):
            connector_column = self_model.get_connector_column(mapped)
            if connector_column is None:
                raise EntityError("No connector key found between %s and %s." % (self_model.__name__, mapped.__name__))

            self.builder.add_to_data("join", mapped.get_table_name())
            if exprs:
                self.builder.data["on"] = exprs
            else:
                exec("from %s import %s" % (self_model.__module__, self_model.__name__), globals(), globals())
                exec("from %s import %s" % (mapped.__module__, mapped.__name__), globals(), globals())
                expr = eval(
                    "%s.%s == %s.%s"
                    % (
                        self_model.__name__,
                        connector_column.variable_name,
                        mapped.__name__,
                        mapped.get_primary_key_column().variable_name,
                    )
                )
                self.builder.add_to_data("on", expr)
        elif isinstance(mapped, Subquery):
            if not exprs:
                raise EntityError("Cannot join subquery without connector expressions.")
            self.builder.add_to_data("join", mapped.full_sql)
            self.builder.data["on"] = exprs

        return self

    def filter(self, *exprs: Any) -> Query:
        """Exerts a series of valid comparison expressions as filtering criteria to the current instance.

        E.g.::

            session.query(User, email_address, last_visited)
                .filter(User.first_name == "Nima", User.age > 28)
                .first()

        Any number of criteria may be specified as separated by a comma.

        .. seealso::

            :meth:`.query.Query.filter_by` - filter on keyword arguments as criteria.
        """
        for expr in exprs:
            self.builder.add_to_data("where", expr)

        return self

    def filter_by(self, **kwcrts: Any) -> Query:
        """Exerts a series of keyword arguments as filtering criteria to the current instance.

        E.g.::

            session.query(Employee, full_name, salary, tenure)
                .filter_by(department="Information Technologies", status="active")
                .all()

        Any number of criteria may be specified as separated by a comma.

        .. seealso::

            :meth:`.query.Query.filter` - filter on valid comparison expressions as criteria.
        """
        model = self.mapped_class or self.entities[0].model
        exec("from %s import %s" % (model.__module__, model.__name__), globals(), globals())
        criteria = [eval("%s.%s == '%s'" % (model.__name__, key, val)) for key, val in kwcrts.items()]
        return self.filter(*criteria)

    def group_by(self, *columns: Union[BaseColumn, str]) -> Query:
        """Adds SQL `GROUP BY` constraint on the current query with the given columns."""
        column_names = [
            column.compound_variable_name if isinstance(column, BaseColumn) else column for column in columns
        ]
        if column_names:
            self.builder.data["group_by"] = column_names

        return self

    def having(self, expr: Any) -> Query:
        """Adds SQL `HAVING` constraint on the current query with the given direct Python expression."""
        self.builder.add_to_data("having", expr)
        return self

    def order_by(self, *columns: Union[BaseColumn, str]) -> Query:
        """Adds SQL `ORDER BY` constraint on the current query with the given columns."""
        column_names = [
            column.compound_variable_name if isinstance(column, BaseColumn) else column for column in columns
        ]
        if column_names:
            self.builder.data["order_by"] = column_names

        return self

    def limit(self, value: int) -> Query:
        """Adds SQL `LIMIT` constraint on the current query with the given limit value."""
        self.builder.add_to_data("limit", "%d" % value)
        return self

    def offset(self, value: int) -> Query:
        """Adds SQL `OFFSET` constraint on the current query with the given offset value."""
        self.builder.add_to_data("offset", "%d" % value)
        return self

    def slice(self, start: int, stop: int) -> Query:
        """Adds slice dampers on the current query, i.e. `start` and `stop`."""
        return self.limit(stop - start).offset(start)

    def desc(self) -> Query:
        """Adds SQL `DESC` constraint to the current query. Raises an exception if the query is not ordered."""
        if "order_by" not in self.builder.data or not self.builder.data["order_by"]:
            raise MethodChainingError("Cannot use `desc` without `order_by`.")

        self.builder.data["desc"] = []

        return self

    def distinct(self) -> Query:
        """Adds SQL `DISTINCT` constraint to the current query."""
        self.builder.data["distinct"] = []
        return self

    def subquery(self) -> Subquery:
        """Gets a subquery whose inner SQL is that of the current query with the selected column names."""
        column_names = [s.split(", ")[1] for s in self.builder.data["select"]]
        return Subquery(self._sql, column_names)

    def get(self, **kwargs: Any) -> Optional[Record]:
        """Gets the result by given criteria, e.g. by primary key. Raises an exception if more than one result is
        found.
        """
        if "where" in self.builder.data and self.builder.data["where"]:
            raise MethodChainingError("Cannot use `get` after `filter` or `filter_by`.")

        query_set = self.filter_by(**kwargs).all()
        if len(query_set) > 1:
            raise MultipleResultsFound

        return query_set[0] if query_set else None

    def all(self) -> QuerySet:
        """Gets all results."""
        engine = AbstractEngine.active_instance
        try:
            engine.cursor.execute(self._sql)
            col_names = [col[0] for col in engine.cursor.description]
            results = [Record(dict(zip(col_names, row)), self) for row in engine.cursor.fetchall()]
        except pyodbc.DatabaseError:
            raise QueryFormatError

        return QuerySet(results)

    def first(self) -> Optional[Record]:
        """Gets the first result."""
        query_set = self.limit(1).all()
        return query_set[0]

    def one_or_none(self) -> Optional[Record]:
        """Gets the result or nothing if does not exist. Raises an exception if more than one result is found."""
        query_set = self.all()
        if len(query_set) > 1:
            raise MultipleResultsFound

        return query_set[0] if query_set else None

    def exists(self) -> bool:
        """Whether or not there are any results."""
        return bool(self.first())

    def count(self) -> int:
        """Gets the number of rows in the current queryset."""
        results = self.all()

        return len(results)

    def update(self, **fields_values) -> None:
        """Two ways of updates:

            1. user = session.query(User).filter(User.username == "nbavari").first()
            user.age += 1

            2. session.query(User).filter(User.username == "nbavari").update(**field_values)
        You have to put `session.save()` after both to persist it.
        """
        self.builder.data["update"] = self.builder.data.pop("select")
        self.builder.data["set"] = list(fields_values.items())
        if "limit" in self.builder.data:
            del self.builder.data["limit"]

    def delete(self) -> None:
        """Example usage:

            session.query(User).filter(User.username == "nbavari").delete()
            session.save()
        This will delete the user with the username "nbavari".
        """
        self.builder.data["delete"] = self.builder.data.pop("select")
        if "limit" in self.builder.data:
            del self.builder.data["limit"]



from .column import Column
from .db_engine import DBEngine
from .db_session import DBSession
from .fkey import CASCADE, ForeignKey
from .model import Model
from .types import Date, DateTime, Float, Integer, Serial, String, Time

__version__ = "1.0.1"

__all__ = [
    "CASCADE",
    "Column",
    "Date",
    "DateTime",
    "DBEngine",
    "DBSession",
    "ForeignKey",
    "Model",
    "Float",
    "Integer",
    "Serial",
    "String",
    "Time",
    "__version__",
]



import unittest

from EnORM.column import VirtualField
from EnORM.exceptions import EntityError
from EnORM.query import Query, Subquery

from .defs import Human

Human.alias = None
sq = Query(Human, Human.full_name, Human.age).filter(Human.age >= 30).subquery()


class TestSubquery(unittest.TestCase):
    def test_subquery_type(self) -> None:
        self.assertIsInstance(sq, Subquery)

    def test_subquery_existing_fields(self) -> None:
        virtual_fld = sq.full_name
        self.assertIsInstance(virtual_fld, VirtualField)
        self.assertEqual(virtual_fld.variable_name, "full_name")
        self.assertEqual(virtual_fld.view_name, sq.view_name)

    def test_subquery_nonexisting_fields(self) -> None:
        with self.assertRaises(EntityError):
            _ = sq.id

    def test_subquery_sql(self) -> None:
        self.assertEqual(sq.inner_sql, "SELECT humans.full_name, humans.age FROM humans WHERE humans.age >= 30")
        self.assertEqual(
            sq.full_sql, "(SELECT humans.full_name, humans.age FROM humans WHERE humans.age >= 30) AS %s" % sq.view_name
        )



"""Contains :class:`.pool.ConnectionPool`."""

from queue import Queue
from threading import Lock
from typing import Optional

import pyodbc


class ConnectionPool:
    """Thread-safe connection pool for managing database connections.

    Uses a FIFO queue to manage connections. Implements lazy initialization of connections.

    :param conn_str:    database location, along with auth params
    :param pool_size:   size of the connection pool.
    """

    def __init__(self, conn_str: str, pool_size: int) -> None:
        self.conn_str = conn_str
        self.pool = Queue(maxsize=pool_size)
        self.lock = Lock()

    def acquire(self, timeout: Optional[float] = None) -> pyodbc.Connection:
        """Acquires a connection from the pool."""
        with self.lock:
            if self.pool.empty():
                return pyodbc.connect(self.conn_str)
        return self.pool.get(timeout=timeout)

    def release(self, conn: pyodbc.Connection) -> None:
        """Releases a connection back to the pool."""
        self.pool.put(conn)

    def close_all(self) -> None:
        """Closes all connections in the pool."""
        while not self.pool.empty():
            conn = self.pool.get()
            conn.close()



import unittest

from EnORM.column import Scalar
from EnORM.functions import (
    char_length,
    concat,
    current_date,
    current_time,
    current_timestamp,
    current_user,
    local_time,
    local_timestamp,
    now,
    random,
    session_user,
    user,
)
from EnORM.query import Query

from .defs import Human


class TestFunctions(unittest.TestCase):
    def test_char_length(self) -> None:
        Human.alias = None
        long_names_query = Query(Human).filter(char_length(Human.full_name) > 15)
        self.assertEqual(str(long_names_query), "SELECT humans.* FROM humans WHERE CHAR_LENGTH(humans.full_name) > 15")
        bare_ex = char_length("example")
        self.assertIsInstance(bare_ex, Scalar)
        self.assertEqual(bare_ex.compound_variable_name, "CHAR_LENGTH(example)")

    def test_concat(self) -> None:
        res = concat("this", " is just", " a test", ".")
        self.assertEqual(res, "this is just a test.")

    def test_current_date(self) -> None:
        res = current_date()
        self.assertIsInstance(res, Scalar)
        self.assertEqual(res.compound_variable_name, "CURRENT_DATE")

    def test_current_time(self) -> None:
        res_wo_prec = current_time()
        res_w_prec = current_time(6)
        self.assertIsInstance(res_wo_prec, Scalar)
        self.assertIsInstance(res_w_prec, Scalar)
        self.assertEqual(res_wo_prec.compound_variable_name, "CURRENT_TIME")
        self.assertEqual(res_w_prec.compound_variable_name, "CURRENT_TIME(6)")

    def test_current_timestamp(self) -> None:
        res_wo_prec = current_timestamp()
        res_w_prec = current_timestamp(6)
        self.assertIsInstance(res_wo_prec, Scalar)
        self.assertIsInstance(res_w_prec, Scalar)
        self.assertEqual(res_wo_prec.compound_variable_name, "CURRENT_TIMESTAMP")
        self.assertEqual(res_w_prec.compound_variable_name, "CURRENT_TIMESTAMP(6)")

    def test_current_user(self) -> None:
        res = current_user()
        self.assertIsInstance(res, Scalar)
        self.assertEqual(res.compound_variable_name, "CURRENT_USER")

    def test_local_time(self) -> None:
        res_wo_prec = local_time()
        res_w_prec = local_time(4)
        self.assertIsInstance(res_wo_prec, Scalar)
        self.assertIsInstance(res_w_prec, Scalar)
        self.assertEqual(res_wo_prec.compound_variable_name, "LOCALTIME")
        self.assertEqual(res_w_prec.compound_variable_name, "LOCALTIME(4)")

    def test_local_timestamp(self) -> None:
        res = local_timestamp()
        self.assertIsInstance(res, Scalar)
        self.assertEqual(res.compound_variable_name, "LOCALTIMESTAMP")

    def test_now(self) -> None:
        res = now()
        self.assertIsInstance(res, Scalar)
        self.assertEqual(res.compound_variable_name, "NOW()")

    def test_random(self) -> None:
        res = random()
        self.assertIsInstance(res, Scalar)
        self.assertEqual(res.compound_variable_name, "RANDOM()")

    def test_session_user(self) -> None:
        res = session_user()
        self.assertIsInstance(res, Scalar)
        self.assertEqual(res.compound_variable_name, "SESSION_USER")

    def test_user(self) -> None:
        res = user()
        self.assertIsInstance(res, Scalar)
        self.assertEqual(res.compound_variable_name, "USER()")



"""Contains definitions for custom exceptions."""

from typing import Any, Optional, Type


class IncompatibleArgument(TypeError):
    """Raised when parameters of incompatible values or incompatible types are used together."""


class EntityError(ValueError):
    """Raised when entities (column-like objects) are not used correctly."""


class MethodChainingError(ValueError):
    """Raised when a wrong sequence of methods are chained together on :class:`.query.Query`."""


class Fixed(Exception):
    """Base class for all messaged exceptions."""

    message: Optional[str] = None

    def __init__(self, *args: Any) -> None:
        super().__init__(self.message, *args)


class ValueOutOfBound(Fixed):
    """Raised when a value of an EnORM type is out of bounds."""

    def __init__(self, type_name: str) -> None:
        self.message = "`%s` value is out of bounds." % type_name
        super().__init__()


class FieldNotExist(Fixed):
    """Raised when the inquired field does not exist in a multi-field object."""

    def __init__(self, field_name: str) -> None:
        self.message = "Field name %s not exists." % field_name
        super().__init__()


class WrongFieldType(Fixed):
    """Raised when the value of a column is of another type than is declared."""

    def __init__(self, field_name: str, exp_type: Type, actual_type: Type) -> None:
        self.message = "Field %s expected %s but got %s." % (
            field_name,
            exp_type,
            actual_type,
        )
        super().__init__()


class MissingRequiredField(Fixed):
    """Raised when a new row is instantiated without a required field."""

    def __init__(self, field_name: str) -> None:
        self.message = "Cannot call without field %s." % field_name
        super().__init__()


class QueryFormatError(Fixed):
    """Raised when a query is formatted incorrectly."""

    message = "Wrong query format."


class MultipleResultsFound(Fixed):
    """Raised when multiple results are found while a single result is expected."""

    message = "Multiple results found."


class OrphanColumn(Fixed):
    """Raised when a column is instantiated outside a model definition."""

    message = "Cannot initialize `Column` outside a `Model`."



import unittest

from EnORM import CASCADE, Column, Float, ForeignKey, Integer, Model, Serial, String
from EnORM.exceptions import FieldNotExist, OrphanColumn
from EnORM.query import Subquery

from .defs import Human, Pet

MODEL_CLS = Pet
FOREIGN_MAPPED = Human

MODEL_INSTANCE = MODEL_CLS(name="Sakura", age=5, owner_id=Serial(21))
PERSON = FOREIGN_MAPPED(id=Serial(21), full_name="Nima Bavari Goudarzi", age=32)


class TestModel(unittest.TestCase):
    def test_model_type(self) -> None:
        self.assertIsInstance(MODEL_INSTANCE, Model)
        self.assertIsInstance(MODEL_INSTANCE, MODEL_CLS)

    def test_model_fields(self) -> None:
        self.assertDictEqual(
            MODEL_CLS.get_fields(), {"name": MODEL_CLS.name, "age": MODEL_CLS.age, "owner_id": MODEL_CLS.owner_id}
        )
        self.assertEqual(MODEL_CLS.name.type, String)
        self.assertEqual(MODEL_CLS.age.type, Integer)

    def test_model_table_name(self) -> None:
        self.assertEqual(MODEL_CLS.get_table_name(), "pets")

    def test_model_primary_key(self) -> None:
        self.assertIsNone(MODEL_CLS.get_primary_key_column())

    def test_model_connector_key(self) -> None:
        FOREIGN_MAPPED.alias = None
        self.assertEqual(MODEL_CLS.get_connector_column(FOREIGN_MAPPED), MODEL_CLS.owner_id)
        self.assertIsNone(FOREIGN_MAPPED.get_connector_column(MODEL_CLS))

    def test_model_existing_compound_attr(self) -> None:
        FOREIGN_MAPPED.alias = None
        self.assertIsInstance(PERSON.pets, Subquery)
        self.assertEqual(
            PERSON.pets.inner_sql,
            "SELECT pets.* FROM pets JOIN humans ON pets.owner_id = humans.id WHERE pets.owner_id = %s" % PERSON.id,
        )

    def test_model_nonexisting_compound_attr(self) -> None:
        with self.assertRaises(FieldNotExist):
            _ = PERSON.tots

    def test_model_label(self) -> None:
        FOREIGN_MAPPED.alias = None
        table_as = FOREIGN_MAPPED.label("h")
        self.assertIsInstance(table_as, type)
        self.assertEqual(table_as, FOREIGN_MAPPED)
        self.assertEqual(table_as.alias, "h")


class TestColumn(unittest.TestCase):
    def test_column_state_outside_model_context(self) -> None:
        c = Column(Float)
        with self.assertRaises(OrphanColumn):
            _ = c.model

    def test_column_attr_vals_with_model_context(self) -> None:
        self.assertEqual(MODEL_INSTANCE.name, "Sakura")
        self.assertEqual(MODEL_INSTANCE.age, 5)

        self.assertEqual(MODEL_CLS.name.model, Pet)
        self.assertEqual(MODEL_CLS.name.variable_name, "name")
        self.assertEqual(MODEL_CLS.name.view_name, "pets")

        self.assertEqual(MODEL_CLS.age.model, Pet)
        self.assertEqual(MODEL_CLS.age.variable_name, "age")
        self.assertEqual(MODEL_CLS.age.view_name, "pets")

        self.assertEqual(MODEL_CLS.owner_id.model, Pet)
        self.assertEqual(MODEL_CLS.owner_id.variable_name, "owner_id")
        self.assertEqual(MODEL_CLS.owner_id.view_name, "pets")

    def test_column_attr_types_with_model_context(self) -> None:
        self.assertEqual(MODEL_CLS.name.type, str)
        self.assertEqual(MODEL_CLS.age.type, int)

    def test_column_label(self) -> None:
        column_as = MODEL_CLS.age.label("pet_age")
        self.assertIsInstance(column_as, Column)
        self.assertEqual(column_as, MODEL_CLS.age)
        self.assertEqual(column_as.alias, "pet_age")


class TestForeignKey(unittest.TestCase):
    def test_foreign_key(self):
        FOREIGN_MAPPED.alias = None
        self.assertIsInstance(MODEL_CLS.owner_id.rel, ForeignKey)
        self.assertIs(MODEL_CLS.owner_id.rel.foreign_model, FOREIGN_MAPPED)
        self.assertIs(MODEL_CLS.owner_id.rel.reverse_name, "pets")
        self.assertIs(MODEL_CLS.owner_id.rel.on_delete, CASCADE)



import unittest

from EnORM.db_engine import AbstractEngine
from EnORM.exceptions import EntityError, FieldNotExist, MethodChainingError
from EnORM.functions import count
from EnORM.query import Query, QuerySet, Record, Subquery

from .defs import FakeEngine, Human, Pet


class TestQuery(unittest.TestCase):
    def test_query_init(self) -> None:
        Human.alias = None
        q = Query(Human, Human.id, Human.full_name)
        self.assertSequenceEqual(q.entities, (Human, Human.id, Human.full_name))
        self.assertEqual(str(q), "SELECT humans.id, humans.full_name FROM humans")

    def test_query_join_model_without_connector(self) -> None:
        Human.alias = None
        with self.assertRaises(EntityError):
            _ = Query(Human, Human.id, Human.full_name).join(Pet)

    def test_query_join_model_with_connector_without_exprs(self) -> None:
        Human.alias = None
        q = Query(Pet, Pet.age).join(Human)
        self.assertEqual(str(q), "SELECT pets.age FROM pets JOIN humans ON pets.owner_id = humans.id")

    def test_query_join_model_with_connector_with_exprs(self) -> None:
        Human.alias = None
        Pet.age.alias = None
        q = Query(Pet, Pet.age).join(Human, Human.id == Pet.owner_id)
        self.assertEqual(str(q), "SELECT pets.age FROM pets JOIN humans ON humans.id = pets.owner_id")

    def test_query_join_subquery_without_exprs(self) -> None:
        Human.alias = None
        sq = Query(Human, Human.id, Human.full_name).subquery()
        with self.assertRaises(EntityError):
            _ = Query(Pet, Pet.name, Pet.age).join(sq)

    def test_query_join_subquery_with_exprs(self) -> None:
        Human.alias = None
        sq = Query(Human, Human.id, Human.full_name).subquery()
        q = Query(Pet, Pet.name, Pet.age).join(sq, sq.full_name == Pet.name)
        sq_view_name = sq.view_name
        self.assertEqual(
            str(q),
            "SELECT pets.name, pets.age FROM pets JOIN (SELECT humans.id, humans.full_name FROM humans) AS %s ON "
            "%s.full_name = pets.name" % (sq_view_name, sq_view_name),
        )

    def test_query_filter(self) -> None:
        Human.alias = None
        q = Query(Human, Human.full_name).filter(Human.age == 44, Human.id > 15)
        self.assertEqual(str(q), "SELECT humans.full_name FROM humans WHERE humans.age = 44 AND humans.id > 15")

    def test_query_filter_by(self) -> None:
        Human.alias = None
        q = Query(Human, Human.full_name).filter_by(age=24, gender="M")
        self.assertEqual(str(q), "SELECT humans.full_name FROM humans WHERE humans.age = 24 AND humans.gender = M")

    def test_query_group_by(self) -> None:
        Human.alias = None
        q = Query(Human, Human.id, Human.full_name).group_by(Human.age)
        self.assertEqual(str(q), "SELECT humans.id, humans.full_name FROM humans GROUP BY humans.age")

    def test_query_group_by_agg(self) -> None:
        Human.alias = None
        q = Query(Human, Human.id, count(Human.age).label("age_count")).group_by("age_count")
        self.assertEqual(str(q), "SELECT humans.id, COUNT(humans.age) AS age_count FROM humans GROUP BY age_count")
        q2 = Query(Human, Human.id, count(Human.age).label("age_count")).group_by(Human.age)
        self.assertEqual(str(q2), "SELECT humans.id, COUNT(humans.age) AS age_count FROM humans GROUP BY humans.age")

    def test_query_having(self) -> None:
        Human.alias = None
        q = Query(Human, Human.id, count(Human.age).label("age_count")).having(count(Human.age) > 5)
        self.assertEqual(str(q), "SELECT humans.id, COUNT(humans.age) AS age_count FROM humans HAVING humans.age > 5")

    def test_query_order_by_labelless(self) -> None:
        Human.alias = None
        Human.full_name.alias = None
        q = Query(Human, Human.id, Human.full_name).order_by(Human.age)
        self.assertEqual(str(q), "SELECT humans.id, humans.full_name FROM humans ORDER BY humans.age")

    def test_query_order_by_labeled(self) -> None:
        Human.alias = None
        Human.full_name.aggs = []
        q = Query(Human, Human.id, Human.full_name.label("fname")).order_by("fname")
        self.assertEqual(str(q), "SELECT humans.id, humans.full_name AS fname FROM humans ORDER BY fname")

    def test_query_limit(self) -> None:
        Human.alias = None
        q = Query(Human, Human.id, Human.full_name).limit(20)
        self.assertEqual(str(q), "SELECT humans.id, humans.full_name FROM humans LIMIT 20")

    def test_query_offset(self) -> None:
        Human.alias = None
        q = Query(Human, Human.id, Human.full_name).offset(10)
        self.assertEqual(str(q), "SELECT humans.id, humans.full_name FROM humans OFFSET 10")

    def test_query_slice(self) -> None:
        Human.alias = None
        q = Query(Human, Human.id, Human.full_name).slice(12, 16)
        self.assertEqual(str(q), "SELECT humans.id, humans.full_name FROM humans LIMIT 4 OFFSET 12")

    def test_query_desc(self) -> None:
        Human.alias = None
        with self.assertRaises(MethodChainingError):
            _ = Query(Human, Human.id, Human.full_name).desc()
        q = Query(Human, Human.id, Human.full_name).order_by(Human.age).desc()
        self.assertEqual(str(q), "SELECT humans.id, humans.full_name FROM humans ORDER BY humans.age DESC")

    def test_query_distinct(self) -> None:
        Human.alias = None
        q = Query(Human, Human.id, Human.full_name).distinct()
        self.assertEqual(str(q), "SELECT DISTINCT humans.id, humans.full_name FROM humans")

    def test_query_subquery(self) -> None:
        Human.alias = None
        q = Query(Human, Human.id, Human.full_name).filter(Human.age == 30)
        sq = q.subquery()
        self.assertIsInstance(sq, Subquery)
        self.assertEqual(str(sq), str(q))
        self.assertListEqual(sq.column_names, ["*", "id", "full_name"])

    def test_query_update_direct(self) -> None:
        Human.alias = None
        q = Query(Human, Human.id, Human.full_name).filter(Human.age == 30)
        q.update(age=20)
        self.assertEqual(str(q), "UPDATE humans SET humans.age = 20 WHERE humans.age = 30")

    def test_query_update_indirect(self) -> None:
        AbstractEngine.active_instance = FakeEngine("blah blah blah")
        Human.alias = None
        q = Query(Human, Human.id, Human.full_name).filter(Human.age == 30)
        result = q.first()
        result.age += 1
        self.assertEqual(
            str(q),
            "UPDATE humans SET humans.age = 31 WHERE humans.age = 30 AND humans.id = 17 AND "
            "humans.full_name = Jacques Trate AND humans.age = 30",
        )
        AbstractEngine.active_instance = None

    def test_query_delete(self) -> None:
        Human.alias = None
        q = Query(Human, Human.id, Human.full_name).filter(Human.age == 30)
        q.delete()
        self.assertEqual(str(q), "DELETE FROM humans WHERE humans.age = 30")

    def test_query_execs(self) -> None:
        AbstractEngine.active_instance = FakeEngine("blah blah blah")
        Human.alias = None
        q = Query(Human, Human.id, Human.full_name).filter(Human.age == 30)
        self.assertEqual(q.count(), 2)
        self.assertTrue(q.exists())
        AbstractEngine.active_instance = None

    def test_query_record_identity(self) -> None:
        AbstractEngine.active_instance = FakeEngine("blah blah blah")
        Human.alias = None
        q = Query(Human, Human.id, Human.full_name, Human.age).filter(Human.age == 30)
        res = q.first()
        self.assertIsInstance(res, Record)
        self.assertDictEqual(res.dct, {"id": 17, "full_name": "Jacques Trate", "age": 30})
        self.assertEqual(res.query, q)

    def test_query_record_nonexistent_attr(self) -> None:
        AbstractEngine.active_instance = FakeEngine("blah blah blah")
        Human.alias = None
        res = Query(Human, Human.id, Human.full_name).filter(Human.age == 30).first()
        with self.assertRaises(FieldNotExist):
            _ = res.phone_number
        AbstractEngine.active_instance = None

    def test_query_record_existent_attr(self) -> None:
        AbstractEngine.active_instance = FakeEngine("blah blah blah")
        Human.alias = None
        res = Query(Human, Human.id, Human.full_name, Human.age).filter(Human.age == 30).first()
        self.assertEqual(res.age, 30)
        AbstractEngine.active_instance = None

    def test_query_record_reverse_name_row(self) -> None:
        AbstractEngine.active_instance = FakeEngine("blah blah blah")
        Human.alias = None
        res = Query(Human).filter(Human.age == 30).first()
        self.assertEqual(
            res.pets.inner_sql,
            "SELECT pets.* FROM pets JOIN humans ON pets.owner_id = humans.id WHERE pets.owner_id = 17",
        )
        AbstractEngine.active_instance = None

    def test_query_record_reverse_name_not_row(self) -> None:
        AbstractEngine.active_instance = FakeEngine("blah blah blah")
        Human.alias = None
        res = Query(Human, Human.id, Human.full_name).filter(Human.age == 30).first()
        with self.assertRaises(FieldNotExist):
            _ = res.pets
        AbstractEngine.active_instance = None

    def test_query_queryset(self) -> None:
        AbstractEngine.active_instance = FakeEngine("blah blah blah")
        Human.alias = None
        q = Query(Human, Human.id, Human.full_name, Human.age).filter(Human.age == 30)
        res = q.all()
        self.assertIsInstance(res, QuerySet)
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[1], Record)
        slc = res[1:]
        self.assertIsInstance(slc, list)
        self.assertEqual(len(slc), 1)
        member = slc[0]
        self.assertIsInstance(member, Record)
        self.assertDictEqual(member.dct, {"id": 34, "full_name": "Joanna Males", "age": 30})
        self.assertEqual(member.query, q)
        AbstractEngine.active_instance = None



"""Contains ORM functions, both aggregate and non-aggregate ones."""

from typing import Optional, Union

from .column import BaseColumn, Scalar


def agg(field: BaseColumn, name_in_sql: str) -> BaseColumn:
    """Describes the basis for all aggregate functions."""
    field.aggs.append(name_in_sql)
    return field


def count(field: BaseColumn) -> BaseColumn:
    """Describes the SQL aggregate function `COUNT`."""
    return agg(field, "COUNT")


def sum_(field: BaseColumn) -> BaseColumn:
    """Describes the SQL aggregate function `SUM`."""
    return agg(field, "SUM")


def avg(field: BaseColumn) -> BaseColumn:
    """Describes the SQL aggregate function `AVG`."""
    return agg(field, "AVG")


def min_(field: BaseColumn) -> BaseColumn:
    """Describes the SQL aggregate function `MIN`."""
    return agg(field, "MIN")


def max_(field: BaseColumn) -> BaseColumn:
    """Describes the SQL aggregate function `MAX`."""
    return agg(field, "MAX")


def char_length(c: Union[BaseColumn, str]) -> Scalar:
    """Describes the SQL function `CHAR_LENGTH`."""
    full_field_name = c
    if isinstance(c, BaseColumn):
        full_field_name = ".".join(c.compound_variable_name.split(", "))
    return Scalar("CHAR_LENGTH(%s)" % full_field_name)


def concat(*parts: str) -> str:
    """Concatenates multiple strings into one."""
    return "".join(parts)


def current_date() -> Scalar:
    """Describes the SQL function `CURRENT_DATE`."""
    return Scalar("CURRENT_DATE")


def current_time(precision: Optional[int] = None) -> Scalar:
    """Describes the SQL function `CURRENT_TIME`."""
    result = "CURRENT_TIME"
    if precision is not None:
        result += "(%d)" % precision
    return Scalar(result)


def current_timestamp(precision: Optional[int] = None) -> Scalar:
    """Describes the SQL function `CURRENT_TIMESTAMP`."""
    result = "CURRENT_TIMESTAMP"
    if precision is not None:
        result += "(%d)" % precision
    return Scalar(result)


def current_user() -> Scalar:
    """Describes the SQL function `CURRENT_USER`."""
    return Scalar("CURRENT_USER")


def local_time(precision: Optional[int] = None) -> Scalar:
    """Describes the SQL function `LOCALTIME`."""
    result = "LOCALTIME"
    if precision is not None:
        result += "(%d)" % precision
    return Scalar(result)


def local_timestamp() -> Scalar:
    """Describes the SQL function `LOCALTIMESTAMP`."""
    return Scalar("LOCALTIMESTAMP")


def now() -> Scalar:
    """Describes the SQL function `NOW`."""
    return Scalar("NOW()")


def random() -> Scalar:
    """Describes the SQL function `RANDOM`."""
    return Scalar("RANDOM()")


def session_user() -> Scalar:
    """Describes the SQL function `SESSION_USER`."""
    return Scalar("SESSION_USER")


def user() -> Scalar:
    """Describes the SQL function `USER`."""
    return Scalar("USER()")



import unittest

from EnORM import Column, DBSession, Model, Serial, String
from EnORM.query import Query

from .defs import FakeEngine


class Order(Model):
    id = Column(Serial, primary_key=True)
    country = Column(String, 30, nullable=False)
    province = Column(String)
    city = Column(String, 30, nullable=False)


class TestDBSession(unittest.TestCase):
    def setUp(self) -> None:
        self.e1 = FakeEngine("remote DB connection string")
        self.e2 = FakeEngine("another DB connection string")
        self.sess1 = DBSession(self.e1)
        self.sess2 = DBSession(self.e2)

    def test_pending_model_definitions_without_session(self) -> None:
        self.assertListEqual(Model.model_definition_sqls, [])

    def test_session_singularity(self) -> None:
        self.assertIs(self.sess1, self.sess2)
        self.assertIs(DBSession._instance, self.sess1)
        self.assertIs(DBSession._instance, self.sess2)

    def test_session_engine_active_instance(self) -> None:
        self.assertIs(FakeEngine.active_instance, self.e2)

    def test_pending_model_definitions_with_session(self) -> None:
        self.assertListEqual(self.sess1.engine.conn.executions, [])

    def test_session_context_manager_exit(self) -> None:
        with DBSession(self.e2) as _:
            pass
        self.assertFalse(self.e2.cursor.open)
        self.assertFalse(self.e2.conn.open)

    def test_session_add_and_commit_adds(self) -> None:
        with DBSession(self.e2) as sess:
            order = Order(country="Germany", city="Hamburg")
            sess.add(order)
            self.assertListEqual(sess.persistence_manager.queue, [order])
            self.assertListEqual(list(sess), [order])
            sess.persistence_manager.auto_commit_adds()
            self.assertListEqual(sess.engine.conn.executions, [])
            self.assertListEqual(sess.query_executor.accumulator, [])

    def test_session_query(self) -> None:
        q = self.sess2.query(Order, Order.country).filter(Order.id == 12)
        self.assertIsInstance(q, Query)
        self.assertEqual(str(q), "SELECT orders.country FROM orders WHERE orders.id = 12")
        self.assertListEqual(self.sess2.query_executor.accumulator, [q])

    def test_session_save(self) -> None:
        self.sess2.save()
        self.assertListEqual(self.sess2.engine.conn.executions, [])
        self.assertListEqual(self.sess2.query_executor.accumulator, [])



"""Contains :class:`.db_session.DBSession`."""

from __future__ import annotations

from types import TracebackType
from typing import Any, Iterator, List, Optional, Type, Union

import pyodbc

from .column import BaseColumn
from .db_engine import AbstractEngine
from .model import Model
from .query import Query


class TransactionManager:
    """Used within repository pattern in :class:`.db_session.DBSession` to manage transactions.

    :param engine:  DB engine that the transaction manager uses.
    """

    def __init__(self, engine: AbstractEngine) -> None:
        self.engine = engine

    def commit(self) -> None:
        """Commits the current transaction."""
        self.engine.conn.commit()

    def rollback(self) -> None:
        """Rolls back the current transaction."""
        self.engine.conn.rollback()

    def close(self) -> None:
        """Closes the connection."""
        self.engine.cursor.close()
        self.engine.conn.close()


class PersistenceManager:
    """Used within repository pattern in :class:`.db_session.DBSession` to manage persistence.

    :param engine:  DB engine that the persistence manager uses.
    """

    def __init__(self, engine: AbstractEngine) -> None:
        self.engine = engine
        self.queue: List[Model] = []

    def add(self, obj: Model) -> None:
        """Adds an object to the queue for persistence."""
        self.queue.append(obj)

    def auto_commit_adds(self) -> None:
        """Persists all added objects."""
        for itm in self.queue:
            self.engine.cursor.execute(itm.sql, *itm.attrs.values())

        self.engine.conn.commit()

        self.queue = []


class QueryExecutor:
    """Used within repository pattern in :class:`.db_session.DBSession` to execute queries.

    :param engine:  DB engine that the query executor uses.
    """

    def __init__(self, engine: AbstractEngine) -> None:
        self.engine = engine
        self.accumulator: List[Query] = []

    def query(self, *fields: Union[Type, BaseColumn]) -> Query:
        """Starts a query and returns the query object."""
        query = Query(*fields)
        self.accumulator.append(query)
        return query

    def execute_queries(self) -> None:
        """Executes accumulated queries."""
        for query in self.accumulator:
            self.engine.cursor.execute(query._sql)

        self.engine.conn.commit()

        self.accumulator = []


class DBSession:
    """A singleton DB session class.

    This class implements a singleton DB session to be used to access the database.

    Implements a context manager for a more secure session. The following is an idiomatic usage::

        eng = DBEngine("postgresql://user:secret@localhost:5432/my_db")
        with DBSession(eng) as session:
            pass  # do something with session

    :param engine:  DB engine that the session uses.
    """

    _instance: Optional[DBSession] = None

    def __new__(cls, *args: Any, **kwargs: Any) -> DBSession:
        if not isinstance(cls._instance, cls):
            cls._instance = super().__new__(cls)

        return cls._instance

    def __init__(self, engine: AbstractEngine) -> None:
        self.engine = engine
        AbstractEngine.active_instance = self.engine

        self.transaction_manager = TransactionManager(self.engine)
        self.persistence_manager = PersistenceManager(self.engine)
        self.query_executor = QueryExecutor(self.engine)

        try:
            for sql in Model.model_definition_sqls:
                self.engine.cursor.execute(sql)
        except pyodbc.DatabaseError:
            raise
        finally:
            Model.model_definition_sqls = []

        self.transaction_manager.commit()

    def __enter__(self) -> DBSession:
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        if not self.engine.conn:
            return

        try:
            if exc_type is None:
                self.persistence_manager.auto_commit_adds()
            else:
                self.transaction_manager.rollback()
                raise exc_value
        finally:
            self.transaction_manager.close()
            self.engine.release_connection(self.engine.conn)
            DBSession._instance = None

    def __iter__(self) -> Iterator[Model]:
        yield from self.persistence_manager.queue

    def query(self, *fields: Union[Type, BaseColumn]) -> Query:
        """Starts a query and returns the query object."""
        return self.query_executor.query(*fields)

    def add(self, obj: Model) -> None:
        """Adds a new record to the DB session."""
        self.persistence_manager.add(obj)

    def save(self) -> None:
        """Persists all started queries."""
        self.query_executor.execute_queries()



"""Contains definitions for custom types."""

from datetime import date, datetime, time

from .exceptions import ValueOutOfBound

Integer = int
Date = date
DateTime = datetime
Float = float
String = str
Time = time


class Serial(Integer):
    """ORM representation of serial types in SQL variants. Inherits from :class:`.types.Integer`.

    :params val:    underlying integer value.
    """

    def __init__(self, val: Integer) -> None:
        self.val = val
        if self.val < 1 or self.val > 2147483647:
            raise ValueOutOfBound("Serial")



"""Contains abstract and concrete database engine classes."""

import pyodbc

from .pool import ConnectionPool


class AbstractEngine:
    """Abstract database engine class."""

    active_instance = None


class DBEngine(AbstractEngine):
    """Connection adapter for the :class:`.db_session.DBSession` object.

    This provides a thin wrapper around DB driver API.

    The class keeps record of the most recent active instance as an inner state.

    :param conn_str:    database location, along with auth params
    :param pool_size:   keyword-only. Size of the connection pool.
    """

    def __init__(self, conn_str: str, *, pool_size: int) -> None:
        self.connection_pool = ConnectionPool(conn_str, pool_size)
        self.conn = self.get_connection()
        self.cursor = self.conn.cursor()

    def get_connection(self) -> pyodbc.Connection:
        """Gets a connection from the pool."""
        return self.connection_pool.acquire()

    def release_connection(self, conn: pyodbc.Connection) -> None:
        """Releases a connection back to the pool."""
        self.connection_pool.release(conn)



"""Contains :class:`.model.Model`."""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Type

from .column import Column
from .exceptions import FieldNotExist, MissingRequiredField, WrongFieldType
from .query import Query
from .types import Float, Integer, Serial, String


class SchemaDefinition:
    """SQL schema definition for a user defined model."""

    def __init__(self, model: Type) -> None:
        self.model = model

    def generate_sql(self) -> str:
        """Generates SQL for creating the associated table."""
        fields = self.model.get_fields()
        tablename = self.model.get_table_name()
        return """DROP TABLE %s IF EXISTS; CREATE TABLE %s (%s);""" % (
            tablename,
            tablename,
            ", ".join(self.model.get_type_pref(field, val) for field, val in fields.items()),
        )


class Model:
    """Abstract representer of the database model in Python.

    This class provides the base for defining the structure and behavior of data objects that will be stored in a
    database. The :class:`.model.Model` class can be inherited and used to create custom models with fields, methods,
    and attributes. It supports relationships between different models, enabling data retrieval and manipulation across
    tables.
    """

    __table__ = None
    __TYPES = {
        Integer: "INTEGER",
        String: "VARCHAR",
        Float: "NUMERIC",
        Serial: "SERIAL",
    }

    dep_mapping: Dict[type, List[type]] = {}
    model_definition_sqls: List[str] = []
    alias: Optional[str] = None

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        schema = SchemaDefinition(cls)
        table_creation_sql = schema.generate_sql()
        cls.model_definition_sqls.append(table_creation_sql)

        fields = cls.get_fields()
        for field, val in fields.items():
            Column.objects[id(val)] = {"model": cls, "variable_name": field}
            if val.rel is not None:
                cls.dep_mapping.setdefault(val.rel.foreign_model, []).append(cls)

    def __init__(self, **attrs: Any) -> None:
        self.attrs = attrs

        fields = self.get_fields()
        for key, val in self.attrs.items():
            if key not in fields.keys():
                raise FieldNotExist(key)

            expected_type = fields[key].type
            if not isinstance(val, expected_type):
                raise WrongFieldType(key, expected_type, type(val))

            setattr(self, key, val)

        for field, val in fields.items():
            if field in self.attrs.keys():
                continue

            if field == "id":
                continue

            if val.default:
                setattr(self, field, val.default)
                continue

            if val.nullable:
                setattr(self, field, None)
                continue

            raise MissingRequiredField(field)

    @classmethod
    def get_fields(cls) -> Dict[str, Column]:
        return {key: val for key, val in cls.__dict__.items() if hasattr(val, "rel")}

    @classmethod
    def get_table_name(cls) -> str:
        return cls.__table__ or "%ss" % cls.__qualname__.split(".")[-1].lower()

    @classmethod
    def get_primary_key_column(cls) -> Optional[Column]:
        return next((c for _, c in cls.get_fields().items() if c.primary_key), None)

    @classmethod
    def get_connector_column(cls, mapped: Type) -> Optional[Column]:
        return next(
            (c for _, c in cls.get_fields().items() if c.rel is not None and c.rel.foreign_model == mapped),
            None,
        )

    @classmethod
    def get_type_pref(cls, field: str, val: Column) -> str:
        type_pref_inc = cls.__TYPES[val.type]
        if val.rel is not None:
            pref = "%s %s FOREIGN KEY (%s) REFERENCES %s(%s)" % (
                field,
                type_pref_inc,
                field,
                val.rel.foreign_model.get_table_name(),
                val.rel.foreign_model.get_primary_key_column().variable_name,
            )
            if val.rel.on_delete is not None:
                pref += " ON DELETE CASCADE"
            if val.rel.on_update is not None:
                pref += " ON UPDATE CASCADE"
            return pref

        if val.length is not None:
            type_pref_inc += " (%d)" % val.length
        elif val.type == String:
            type_pref_inc = "TEXT"

        if val.primary_key:
            if val.type == Serial:
                type_pref_inc += " AUTOINCREMENT"
            type_pref_inc += " PRIMARY KEY"

        if val.default:
            type_pref_inc += " DEFAULT %s" % val.default

        if not val.nullable:
            type_pref_inc += " NOT NULL"

        return "%s %s" % (field, type_pref_inc)

    @classmethod
    def label(cls, alias: str) -> Type:
        """Returns the ORM representation for SQL table aliasing."""
        cls.alias = alias
        return cls

    def __getattr__(self, attr: str) -> Any:
        self_model = type(self)
        try:
            fields = self_model.get_fields()
            return fields[attr]
        except KeyError as e:
            for m in self_model.dep_mapping[self_model]:
                connector = m.get_connector_column(self_model)
                if connector.rel.reverse_name == attr:
                    condition_dict = {
                        connector.variable_name: getattr(self, self.get_primary_key_column().variable_name)
                    }
                    return Query(m).join(self_model).filter_by(**condition_dict).subquery()

            raise FieldNotExist(attr) from e

    @property
    def sql(self) -> str:
        """SQL statement for new object creation."""
        return """INSERT INTO %s (%s) VALUES (%s);""" % (
            self.get_table_name(),
            ", ".join(self.attrs.keys()),
            ", ".join("?" for _ in self.attrs),
        )
